==> ksqldb-engine-7.1.9-sources.jar: io/confluent/ksql/engine/RuntimeAssignor.java <==
/*
 * Copyright 2021 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.engine;

import static io.confluent.ksql.util.QueryApplicationId.buildSharedRuntimeId;

import com.google.common.collect.ImmutableMap;
import io.confluent.ksql.name.SourceName;
import io.confluent.ksql.query.QueryId;
import io.confluent.ksql.util.BinPackedPersistentQueryMetadataImpl;
import io.confluent.ksql.util.KsqlConfig;
import io.confluent.ksql.util.PersistentQueryMetadata;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class RuntimeAssignor {
  private final Map<String, Collection<SourceName>> runtimesToSources;
  private final Map<QueryId, String> idToRuntime;

  public RuntimeAssignor(final KsqlConfig config) {
    runtimesToSources = new HashMap<>();
    idToRuntime = new HashMap<>();
    final int runtimes = config.getInt(KsqlConfig.KSQL_SHARED_RUNTIMES_COUNT);
    for (int i = 0; i < runtimes; i++) {
      final String runtime = buildSharedRuntimeId(config, true, i);
      runtimesToSources.put(runtime, new HashSet<>());
    }
  }

  private RuntimeAssignor(final RuntimeAssignor other) {
    this.runtimesToSources = new HashMap<>();
    this.idToRuntime = new HashMap<>(other.idToRuntime);
    for (Map.Entry<String, Collection<SourceName>> runtime
        : other.runtimesToSources.entrySet()) {
      this.runtimesToSources.put(runtime.getKey(), new HashSet<>(runtime.getValue()));
    }
  }

  public RuntimeAssignor createSandbox() {
    return new RuntimeAssignor(this);
  }

  public String getRuntimeAndMaybeAddRuntime(final QueryId queryId,
                                             final Collection<SourceName> sources,
                                             final KsqlConfig config) {
    if (idToRuntime.containsKey(queryId)) {
      return idToRuntime.get(queryId);
    }
    final List<String> possibleRuntimes = runtimesToSources.entrySet()
        .stream()
        .filter(t -> t.getValue().stream().noneMatch(sources::contains))
        .map(Map.Entry::getKey)
        .collect(Collectors.toList());
    final String runtime;
    if (possibleRuntimes.isEmpty()) {
      runtime = makeNewRuntime(config);
    } else {
      runtime = possibleRuntimes.get(Math.abs(queryId.hashCode() % possibleRuntimes.size()));
    }
    runtimesToSources.get(runtime).addAll(sources);
    idToRuntime.put(queryId, runtime);
    return runtime;
  }

  public void dropQuery(final PersistentQueryMetadata queryMetadata) {
    if (queryMetadata instanceof BinPackedPersistentQueryMetadataImpl) {
      runtimesToSources.get(queryMetadata.getQueryApplicationId())
          .removeAll(queryMetadata.getSourceNames());
      idToRuntime.remove(queryMetadata.getQueryId());
    }
  }


  private String makeNewRuntime(final KsqlConfig config) {
    final String runtime = buildSharedRuntimeId(config, true, runtimesToSources.size());
    runtimesToSources.put(runtime, new HashSet<>());
    return runtime;
  }

  public void rebuildAssignment(final Collection<PersistentQueryMetadata> queries) {
    for (PersistentQueryMetadata queryMetadata: queries) {
      if (queryMetadata instanceof BinPackedPersistentQueryMetadataImpl) {
        runtimesToSources.put(queryMetadata.getQueryApplicationId(),
            queryMetadata.getSourceNames());
        idToRuntime.put(queryMetadata.getQueryId(), queryMetadata.getQueryApplicationId());
      }
    }
  }

  public Map<String, Collection<SourceName>> getRuntimesToSources() {
    return ImmutableMap.copyOf(runtimesToSources);
  }

  public Map<QueryId, String> getIdToRuntime() {
    return ImmutableMap.copyOf(idToRuntime);
  }
}

==> ksqldb-engine-7.2.7-sources.jar: io/confluent/ksql/engine/RuntimeAssignor.java <==
/*
 * Copyright 2021 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.engine;

import static io.confluent.ksql.util.QueryApplicationId.buildSharedRuntimeId;

import com.google.common.collect.ImmutableMap;
import io.confluent.ksql.name.SourceName;
import io.confluent.ksql.query.QueryId;
import io.confluent.ksql.util.BinPackedPersistentQueryMetadataImpl;
import io.confluent.ksql.util.KsqlConfig;
import io.confluent.ksql.util.PersistentQueryMetadata;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class RuntimeAssignor {
  private final Map<String, Collection<SourceName>> runtimesToSources;
  private final Map<QueryId, String> idToRuntime;
  private final Logger log = LoggerFactory.getLogger(RuntimeAssignor.class);
  private final int numDefaultRuntimes;

  public RuntimeAssignor(final KsqlConfig config) {
    runtimesToSources = new HashMap<>();
    idToRuntime = new HashMap<>();
    numDefaultRuntimes = config.getInt(KsqlConfig.KSQL_SHARED_RUNTIMES_COUNT);
    for (int i = 0; i < numDefaultRuntimes; i++) {
      final String runtime = buildSharedRuntimeId(config, true, i);
      runtimesToSources.put(runtime, new HashSet<>());
    }
  }

  private RuntimeAssignor(final RuntimeAssignor other) {
    numDefaultRuntimes = other.numDefaultRuntimes;
    this.runtimesToSources = new HashMap<>();
    this.idToRuntime = new HashMap<>(other.idToRuntime);
    for (Map.Entry<String, Collection<SourceName>> runtime
        : other.runtimesToSources.entrySet()) {
      this.runtimesToSources.put(runtime.getKey(), new HashSet<>(runtime.getValue()));
    }
  }

  public RuntimeAssignor createSandbox() {
    return new RuntimeAssignor(this);
  }

  public String getRuntimeAndMaybeAddRuntime(final QueryId queryId,
                                             final Collection<SourceName> sources,
                                             final KsqlConfig config) {
    if (idToRuntime.containsKey(queryId)) {
      return idToRuntime.get(queryId);
    }
    final List<String> possibleRuntimes = runtimesToSources.entrySet()
        .stream()
        .filter(t -> t.getValue().stream().noneMatch(sources::contains))
        .map(Map.Entry::getKey)
        .collect(Collectors.toList());
    final String runtime;
    if (possibleRuntimes.isEmpty()) {
      runtime = makeNewRuntime(config);
    } else {
      runtime = possibleRuntimes.get(Math.abs(queryId.hashCode() % possibleRuntimes.size()));
    }
    runtimesToSources.get(runtime).addAll(sources);
    idToRuntime.put(queryId, runtime);
    log.info("Assigning query {} to runtime {}", queryId, runtime);
    return runtime;
  }

  public void dropQuery(final PersistentQueryMetadata queryMetadata) {
    if (queryMetadata instanceof BinPackedPersistentQueryMetadataImpl) {
      if (idToRuntime.containsKey(queryMetadata.getQueryId())) {
        log.info("Unassigning query {} from runtime {}",
            queryMetadata.getQueryId(),
            idToRuntime.get(queryMetadata.getQueryId()));
      } else {
        log.warn("Dropping an unassigned query {}, this should"
            + " only possible with Gen 1 queries", queryMetadata);
      }
      runtimesToSources.get(queryMetadata.getQueryApplicationId())
          .removeAll(queryMetadata.getSourceNames());
      idToRuntime.remove(queryMetadata.getQueryId());
      if (runtimesToSources.get(queryMetadata.getQueryApplicationId()).isEmpty()
          && runtimesToSources.size() > numDefaultRuntimes) {
        runtimesToSources.remove(queryMetadata.getQueryApplicationId());
        log.info("Removing runtime {} form selection of possible runtimes",
            queryMetadata.getQueryApplicationId());
      }
    }
  }


  private String makeNewRuntime(final KsqlConfig config) {
    final String runtime = buildSharedRuntimeId(config, true, runtimesToSources.size());
    runtimesToSources.put(runtime, new HashSet<>());
    return runtime;
  }

  public void rebuildAssignment(final Collection<PersistentQueryMetadata> queries) {
    final Set<QueryId> gen1Queries = new HashSet<>();
    for (PersistentQueryMetadata queryMetadata: queries) {
      if (queryMetadata instanceof BinPackedPersistentQueryMetadataImpl) {
        if (!runtimesToSources.containsKey(queryMetadata.getQueryApplicationId())) {
          runtimesToSources.put(queryMetadata.getQueryApplicationId(), new HashSet<>());
        }
        runtimesToSources.get(queryMetadata.getQueryApplicationId())
            .addAll(queryMetadata.getSourceNames());
        idToRuntime.put(queryMetadata.getQueryId(), queryMetadata.getQueryApplicationId());
      } else {
        gen1Queries.add(queryMetadata.getQueryId());
      }
    }
    if (!idToRuntime.isEmpty()) {
      log.info("The current assignment of queries to Gen 2 runtimes is: {}",
          idToRuntime.entrySet()
              .stream()
              .map(e -> e.getKey() + "->" + e.getValue())
              .collect(Collectors.joining(", ")));
    } else {
      if (gen1Queries.size() == queries.size()) {
        log.info("There are no queries assigned to Gen 2 runtimes yet.");
      } else {
        log.error("Gen 2 queries are not getting assigned correctly, this should not be possible");
      }
    }
    if (!gen1Queries.isEmpty()) {
      log.info("Currently there are {} queries running on the Gen 1 runtime which are: {}",
          gen1Queries.size(),
          gen1Queries);
    }
  }

  public Map<String, Collection<SourceName>> getRuntimesToSources() {
    return ImmutableMap.copyOf(runtimesToSources);
  }

  public Map<QueryId, String> getIdToRuntime() {
    return ImmutableMap.copyOf(idToRuntime);
  }
}
